package ssw.mj.symtab;

/**
 * MicroJava Symbol Table Objects:
 * Every named object in a program is stored in an <code>Obj</code> node.
 * Every scope has a list of objects declared within it.
 *
 * @author AW
 */
public class Obj {
	public static final int 
		Con = 0, Var = 1, Type = 2, Meth = 3, Prog = 4;
	
	public int kind;     // Con, Var, Type, Meth, Prog
	public String name;
	public Struct type;
	public Obj next;     // next Obj node in this scope
	public int val;      // Con: value;
	public int adr;      // Var, Meth: offset
	public int level;    // Var: declaration level (0..global, 1..local)
	public int nPars;    // Meth: # of parameters
	public Obj locals;   // Meth: list of local variables
	                     // Prog: symbol table of program
	
	public Obj (int kind, String name, Struct type) {
		this.kind = kind; this.name = name; this.type = type;
	}
	
	public boolean equals (Object o) {
		if (this == o) return true;  // same object
		if (!(o instanceof Obj)) return false;
		return equals((Obj) o);
	}

	public boolean equals (Obj o) {
		if (kind != o.kind || !name.equals(o.name) || !type.equals(o.type))
			return false;
		
		switch (kind) {
			case Con: return val == o.val;
			case Var: return adr == o.adr && level == o.level;
			case Meth: return adr == o.adr && nPars == o.nPars && 
			                  equalsCompleteList(locals, o.locals);
			case Prog: return equalsCompleteList(locals, o.locals);
		}
		
		return true;	
	}
	
	public String toString () {
		StringBuffer sb = new StringBuffer();

		switch (kind) {
			case  Con: sb.append("Constant "); break;
			case  Var: 
				if (level == 0) sb.append("global ");
				else sb.append("local ");
				sb.append("Var " + adr + " "); 
				break;
			case Type: sb.append("Type "); break;
			case Meth: sb.append("Method "); break;
			case Prog: sb.append("Program "); break;
		}

		if (kind == Prog) sb.append(name);
		else sb.append(type + " " + name);

		switch (kind) {
			case  Con: 
				sb.append(" = ");
				if (type == Tab.charType) sb.append("'" + (char) val + "'");
				else sb.append(val);
				break;
			case Meth: 
				sb.append("("); sb.append(nPars); sb.append(") at "); sb.append(adr);
				if (locals != null) {
					sb.append("\n");
					for (Obj l = locals; l != null; l = l.next) {
						sb.append(l); sb.append("\n");
					}
				}
				break;
			case Prog:
				if (locals != null) {
					sb.append("\n");
					for (Obj l = locals; l != null; l = l.next) {
						sb.append(l); sb.append("\n");
					}
				}
		}
		
		return sb.toString();
	}
	
	/** Compare complete Obj node list. */
	public static boolean equalsCompleteList (Obj o1, Obj o2) {
		if (o1 == o2) return true;  // same object
		
		while (o1 != null && o1.equals(o2)) {
			o1 = o1.next; o2 = o2.next;
		}
		
		return o1 == null && o2 == null;
	}
}
